#!/usr/bin/env perl
# $Id: testscript.t,v 1.20 2021/03/15 03:10:23 dpchrist Exp $
# By David Paul Christensen dpchrist@holgerdanske.com
# Public Domain.

use strict;
use warnings;

use Capture::Tiny qw(capture);
use Data::Dumper;
use File::Basename;
use File::Slurp;
use File::Spec::Functions;
use Getopt::Long;
use Test::More;

my $visudo      = '/usr/sbin/visudo';

my $base	= basename $0;
my $debug	= undef;
my $verbose	= undef;
my $xtrace	= undef;
my @opt;

my @cases = (
    [ __LINE__,
    "",
    "",
    "Should not modify an empty file\n"],

    [ __LINE__,
    "\n",
    "\n",
    "Single line - should not modify a blank line\n"],
    [ __LINE__,
    " \n",
    " \n",
    "Single line - should not modify a line with one space\n"],

    [ __LINE__,
    "\t\n",
    "\t\n",
    "Single line - should not modify a line with one tab\n"],

    [ __LINE__,
    "# blah blah blah\n",
    "# blah blah blah\n",
    "Single line - should not modify a comment line\n"],
    [ __LINE__,
    "\t# blah blah blah\n",
    "\t# blah blah blah\n",
    "Single line - should not modify an indented comment line\n"],

    [ __LINE__,
    "Defaults secure_path = /usr/local/bin\n",
    "Defaults secure_path = /usr/local/bin\n",
    "Single line - path exists already and is only component\n"],

    [ __LINE__,
    "Defaults secure_path = /usr/local/bins\n",
    "Defaults secure_path = /usr/local/bins:/usr/local/bin\n",
    "Single line - mispelled path exists and is only component\n"],

    [ __LINE__,
    "Defaults secure_path = /sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path = /sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - appends new path\n"],
    [ __LINE__,
    "Defaults\tsecure_path = /sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults\tsecure_path = /sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - tab after 'Defaults' - Appends new path\n"],
    [ __LINE__,
    "Defaults \tsecure_path = /sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults \tsecure_path = /sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - space-Tab after 'Defaults' - Appends new path\n"],
    [ __LINE__,
    "Defaults secure_path= /sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path= /sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - no white space after 'secure_path' - Appends new path\n"],
    [ __LINE__,
    "Defaults secure_path\t= /sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path\t= /sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - tab after 'secure_path' - Appends new path\n"],
    [ __LINE__,
    "Defaults secure_path \t= /sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path \t= /sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - space-Tab after 'secure_path' - Appends new path\n"],
    [ __LINE__,
    "Defaults secure_path\t = /sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path\t = /sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - Tab-Space after 'secure_path' - Appends new path\n"],
    [ __LINE__,
    "Defaults secure_path =/sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path =/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - no white space after '=' - Appends new path\n"],
    [ __LINE__,
    "Defaults secure_path =\t/sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path =\t/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - tab after '=' - Appends new path\n"],
    [ __LINE__,
    "Defaults secure_path = \t/sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path = \t/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - space-Tab after '=' - Appends new path\n"],
    [ __LINE__,
    "Defaults secure_path =\t /sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path =\t /sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Single line - tab-Space after '=' - Appends new path\n"],

    [ __LINE__,
    "Defaults secure_path = /bin:/usr/local/bin:/root/bin\n",
    "Defaults secure_path = /bin:/usr/local/bin:/root/bin\n",
    "Single line - declines to append because path is already in the MIDDLE\n"],
    [ __LINE__,
    "Defaults secure_path = /bin:/root/bin:/usr/local/bin\n",
    "Defaults secure_path = /bin:/root/bin:/usr/local/bin\n",
    "Single line - declines to append because path is already at the END\n"],
    [ __LINE__,
    "Defaults secure_path = /usr/local/bin:/bin:/root/bin\n",
    "Defaults secure_path = /usr/local/bin:/bin:/root/bin\n",
    "Single line - declines to append because path is already at the START\n"],

    [ __LINE__,
    "Defaults secure_path = \\\n\t/sbin:/bin:/usr/sbin:/usr/bin\n",
    "Defaults secure_path = \\\n\t/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin\n",
    "Continue after '=' - appends new path\n"],
    [ __LINE__,
    "Defaults secure_path = \\\n\t/sbin:/bin:/usr/sbin:/usr/bin # blah\n",
    "Defaults secure_path = \\\n\t/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/bin # blah\n",
    "Continue after '=' with comment - appends new path\n"],

    [ __LINE__,
    "Defaults secure_path = /bin\n\tDefaults secure_path = /usr/local/bin:/sbin\n",
    "Defaults secure_path = /bin:/usr/local/bin\n\tDefaults secure_path = /usr/local/bin:/sbin\n",
    "Multiple secure_path lines - appends new path when needed\n"],
);

my %validate_cache;


sub d;
sub dd;
sub validate_syntax;
sub invoke_solution_file;

sub _d { note(join($", (caller(0))[1,2], dd(@_))) if $debug }

sub dd { Data::Dumper->Dump(@_) }

sub invoke_solution_file
{
    my $s	= canonpath shift;

    my $fin	= shift;

    validate_syntax($fin) or die "$fin has invalid syntax";

    (my $fout = "$s-$fin") =~ s/\.in$/.out/;
    if (-e $fout) {
	warn "+ rm -f $fout" if $xtrace;
	unlink $fout;
	delete $validate_cache{$fout} if exists $validate_cache{$fout};
    }

    my @c = ("./$s", $fin);
    warn "+ @c\n" if $xtrace;
    my ($o, $e, $x) = capture { system @c };
    _d [$o], [qw(o)];
    _d [$e], [qw(e)];
    _d [$x], [qw(x)];

    if ($o) {	
	warn "$fout\n" if $verbose;
	write_file($fout, $o);

	ok validate_syntax($fout), "$fout has valid syntax";
    }

    return $o, $e, $x, $fout;
}

sub validate_syntax
{
    my ($f) = @_;

    unless (exists $validate_cache{$f}) {
	my @c = ($visudo, '-c', '-q', '-s', '-f', $f);
	warn "+ @c" if $xtrace;
	my $x = system @c;

	$validate_cache{$f} = !$x;
    }

    return $validate_cache{$f} if exists $validate_cache{$f};
}


### main
{
    GetOptions(
	'debug|d'	=> \$debug,
	'verbose|v'	=> \$verbose,
	'xtrace|x'	=> \$xtrace,
    ) or warn;


    die "Usage: perl $base [-d] [-v] [-x] solution...\n" unless @ARGV;


    ok !validate_syntax($_), "$_ has invalid syntax"
	for glob 'invalid.*.in';


    foreach my $case (@cases) {
	my ($line, $in, $exp, $name) = @$case;

	my $fin = "sudoers.$line.in";
	warn "$fin\n" if $verbose;
	write_file $fin, $in;

	my $fexp = "sudoers.$line.expected";
	warn "$fexp\n" if $verbose;
	write_file $fexp, $exp;

	my $fname = "sudoers.$line.name";
	warn "$fname\n" if $verbose;
	write_file $fname, $name;
    }

    ok validate_syntax($_), "$_ has valid syntax"
	for glob 'sudoers.*.in';


    foreach my $s (@ARGV) {
	_d [$s], [qw(s)];

	foreach my $f (glob 'sudoers.*.in') {
    	    _d [$f], [qw(f)];

	    my ($o, $e, $x) = invoke_solution_file $s, $f;
	    _d [$o], [qw(o)];
	    _d [$e], [qw(e)];
	    _d [$x], [qw(x)];

	    (my $fexp = $f) =~ s/\.in$/.expected/;
	    _d [$fexp], [qw(fexp)];

	    my $expected = read_file $fexp;
	    _d [$expected], [qw(expected)];

	    (my $fname = $f) =~ s/\.in$/.name/;
	    _d [$fname], [qw(fname)];

	    my $name = read_file $fname;
	    chomp($name);
	    _d [$name], [qw(name)];

	    is $o, $expected, "'$s $f' stdout $name";

	    is $e, '', "'$s $f' stderr";

	    is $x, 0, "'$s $f' exit_value $name";
	}
    }
}


done_testing;
